// ============================================================================
//! @file
//! Copyright(c) All Rights Reserved
//!
//! Summary:
// ============================================================================
#include "logger.h"

#include <cstdarg>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <vector>

namespace UT {
//! flag for UT mode
static bool ut_mode = false;
//! holds list of lines currently logged
static std::vector<std::string> ut_lines;
} // namespace UT

// --------------------
//! @brief constructor
Logger::Logger() noexcept = default;

// -------------------
//! @brief write a "====" line with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::start(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, "====", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a line with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::line(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, " ", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write an "ok" line with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::ok(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, "OK", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a "err" line with the given message
//! note: printed even if not verbose
//!
//! @param fmt     the message to write using printf formatting
void Logger::err(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(true, "ERR", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a "bug" line with the given message
//! note: printed even if not verbose
//!
//! @param fmt     the message to write using printf formatting
void Logger::bug(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(true, "BUG", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a "warn" line with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::warn(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, "WARN", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a "dbg" line with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::dbg(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, "DBG", fmt, args);
    va_end(args);
}

// -------------------
//! @brief write a raw line (no tag) with the given message
//!
//! @param fmt     the message to write using printf formatting
void Logger::raw(const char* const fmt, ...) const
{
    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, nullptr, fmt, args);
    va_end(args);
}

// -------------------
//! @brief write an output line with the given message
//!
//! @param lineno  (optional) the current line number for each line printed
//! @param fmt     the message to write
void Logger::output(int lineno, const char* const fmt, ...) const
{
    std::string tag;
    if (lineno == -1) {
        tag = " --     ";
    } else {
        std::stringstream ss;
        tag += " -- ";
        ss << std::right << std::setw(3) << lineno;
        tag += ss.str();
        tag += "]";
    }

    va_list args;
    va_start(args, fmt);
    Logger::_write_line(false, tag.c_str(), fmt, args);
    va_end(args);
}

// -------------------
//! @brief write the given line to stdout
//!
//! @param print_always print even if not verbose
//! @param tag     the initial tag
//! @param fmt     the message to write in fprintf format
//! @param args    (optional) any args required for fmt
void Logger::_write_line(const bool print_always, const char* const tag, const char* const fmt, va_list args) const
{
    if (!print_always && !m_verbose) {
        return;
    }
    char* msg;
    (void)!vasprintf(&msg, fmt, args);

    std::stringstream ss;
    if (tag == nullptr) {
        ss << msg;
    } else {
        ss << std::left << std::setw(4) << tag << " " << msg;
    }
    free(msg);

    // TODO add ability to optionally save to file
    // TODO all are printed to file even if non-verbose

    if (::UT::ut_mode) {
        ::UT::ut_lines.push_back(ss.str());
    } else {
        // coverage: not used in UTs
        std::cout << ss.str() << std::endl; // LCOV_EXCL_LINE
    }
}

// === UT functions from here on

// -------------------
void Logger::ut_set_mode()
{
    ::UT::ut_mode = true;
}

// -------------------
void Logger::ut_clear()
{
    ::UT::ut_lines.clear();
}

// -------------------
auto Logger::ut_lines() -> std::vector<std::string>
{
    return ::UT::ut_lines;
}
